;;; A Common Lisp implementation of the compo protocol Version 2
;;;
;;; Package :compo
;;;
;;; Copyright (C) 2001  Bruno LARTILLOT
;;; 
;;; This program is free software; you can redistribute it and/or
;;; modify it under the terms of the GNU General Public License
;;; as published by the Free Software Foundation; either version 2
;;; of the License, or any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with this program; if not, write to the Free Software
;;; Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

(in-package :compo)

(defclass recursion (context-item) ()
  (:documentation 

"Class
   recursion
Direct Superclasses
   context-item
Purpose
   This context item is responsible of the recursive applying of the realization process
   to the subcompositions of the composition being realized." ) )

(defmethod realize-using-context-item ((context-item recursion) 
                                       (compo compo)
                                       &rest restargs
                                       &key depth &allow-other-keys )

"Primary method
   realize-using-context-item specialized on the recursion type
Behaviour
   The next method is first called.
   1) An argument list is constructed by adding in front of the received &rest argument
      list the composition as a :super-compo key argument and the received depth minus 1,
      provided that a non-null depth has been received, as a :depth key argument. In each
      recursive call to 'realize-using-context-item' as specified in 3), this list is
      augmented with the preceding subcomposition (nil if it is the first) as a
      :prev-compo key argument.
   2) The subcontext list to be used in 3) is the current value of the slot 'subs' in the
      current context if this value is not atomic, or a list of clones of this value, of
      the same length of the compositions's subcomposition list if it is atomic.
   3) The generic function 'realize-using-context-item' is recursively applied to each
      subcomposition using each corresponding subcontext from the subcontext list
      obtained in 2), and with the argument list obtained in 1), as a remaining argument
      list. The recursive calls stop as soon as one of the two lists is reached. 
   4) The composition is given the list of the results of all the calls to
      'realize-using-context-item' as a new subcomposition list."

  (call-next-method)
  (let* ((tmp (append (list :super-compo compo)
                      (and depth (list :depth (- depth 1)))
                      restargs ))
         (new-args (append (list :prev-compo nil) tmp)) )
    (labels ((g (b c)
               (when (and b (consp c))
                 (apply #'realize-using-context-item
                        (make-instance 'realization
                          :context (if (consp b) (car b) (clone b)) )
                        (car c)
                        new-args )
                 (setf new-args (append (list :prev-compo (car c)) tmp))
                 (cons (car c)
                       (g (if (consp b) (cdr b) b) (cdr c)) ) ) ))
      (setf (subs compo) (g (subs context-item) (subs compo))) ) )
  compo )