;;; A Common Lisp implementation of the compo protocol Version 2
;;;
;;; Package :compo
;;;
;;; Copyright (C) 2001  Bruno LARTILLOT
;;; 
;;; This program is free software; you can redistribute it and/or
;;; modify it under the terms of the GNU General Public License
;;; as published by the Free Software Foundation; either version 2
;;; of the License, or any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with this program; if not, write to the Free Software
;;; Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

(in-package :compo)

(defgeneric is-equal (x y)
  (:documentation
"Generic Function
   is-equal (x y)
Arguments
   x and y are two objects to be compared for equality.
Values
   Any value other than nil in case of equality (in some appropriate sense to be defined
   in each primary method), nil otherwise.
Purpose
   Generalization of the Common Lisp function 'equal' to compositions, and more 
   generally to all CLOS objects.
   An 'is-equal' primary method specialized on each class (built-in or user) which
   instances hold slots should be provided. Currently, 'is-equal' has only been
   specialized on the 'cons' built-in class, the 'standard-object' class and the user
   classes defined in the scope of the compo project. Applying 'is-equal' on an instance
   of any other built-in class will not fail, but will return the same result as using
   the standard Common Lisp function 'equal'." ) )

(defmethod is-equal (x y)

"Primary method
   is-equal specialized on 't'
Behaviour
   Equivalent to (equal x y), x and y being the two arguments given to is-equal."

  (equal x y) )

(defmethod is-equal ((x cons) (y cons))

"Primary method
   is-equal specialized on 'cons'
Behaviour
   Let x and y the two arguments given to 'is-equal'.
   If x and y are both lists circular on their car, return the result of recursively
   applying 'is-equal' to the cdr of x and y.
   If x and y are both lists circular on their cdr, return the result of recursively
   applying 'is-equal' to the car of x and y.
   Else return the result of recursively applying 'is-equal' to the car of x and y, and
   to the cdr of x and y."

  (and
   (if (and (eq (car x) x) (eq (car y) y))
     t
     (is-equal (car x) (car y)) )
   (if (and (eq (cdr x) x) (eq (cdr y) y))
     t
     (is-equal (cdr x) (cdr y)) ) ) )

(defmethod is-equal ((x standard-object) (y standard-object))

"Primary method
   is-equal specialized on the 'standard-object' type
Behaviour
   Let x and y the two arguments given to 'is-equal'.
   Return x if the classes of x and y are the same class object, returns nil otherwise.
   A new defined subclass of 'standard-object' is able to specialize the comparison for
   equality process only for the supplementary instance slots it defines by simply
   specializing 'is-equal'."

  (and
   (eq (class-of x) (class-of y))
   x ) )

(defmethod is-equal ((x abstract-compo) (y abstract-compo))

"Primary method
   is-equal specialized on 'abstract-compo'.
Behaviour
   Return x if the next-method returns a non-nil value, and is-equal successively applied
   to the 'name', 'coercer' and 'subs' slots from x and y is true, returns nil otherwise."

  (and
   (call-next-method)
   (is-equal (name x) (name y))
   (is-equal (coercer x) (coercer y))
   (if (eq (subs x) x)
     (eq (subs y) y)
     (is-equal (subs x) (subs y)) )
   x ) )