;;; A Common Lisp implementation of the compo protocol Version 2
;;;
;;; Package :compo
;;;
;;; Copyright (C) 2001  Bruno LARTILLOT
;;; 
;;; This program is free software; you can redistribute it and/or
;;; modify it under the terms of the GNU General Public License
;;; as published by the Free Software Foundation; either version 2
;;; of the License, or any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with this program; if not, write to the Free Software
;;; Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

(in-package :compo)

(defgeneric specialized-normalize (spec class l defaults &key &allow-other-keys)
  (:documentation

"Generic Function
   specialized-normalize (spec class l defaults &key &allow-other-keys)
Arguments
   An instance of a subclass of 'compo' as a specializer. 
   The class object corresponding to the class of the root of the composition currently
   being normalized.
   The parameter list given to the construction of a composition or subcomposition, as
   figuring in the reader stream.
   A remaining key argument list.
Values
   The parameter list normalized in the sense that it can be given to make-instance
   applied on a subclass of 'compo', as a key argument list.
Purpose
   This generic function operates only on the first parameter of the parameter list, and
   appends its result to the normalization of the rest of the parameter list, recursively
   delegated to 'normalize-compo' (see 'normalize-compo'). Each primary method is in
   charge of the recursive call to 'normalize-compo' and of the concatenation of the
   results. Hence, it is allowed to decide how many items in the begining of the
   parameter list constitutes the first parameter (it can be one item alone, a couple
   constituted by a keyword and a value...). For example, the list (\"compo-name\"...)
   will be normalized to (:name \"compo-name\"...), since the presence of a string
   denotes by definition the presence of a composition name. Note that the list
   (:name \"compo-name\"...) will give the same result, since the presence of the
   normalized form of a parameter is always allowed as an entry." ) )

(defmethod specialized-normalize ((spec abstract-compo) class l defaults
                                  &rest restargs
                                  &key &allow-other-keys)

"Primary method
   specialized-normalize specialized on 'compo'
Behaviour
   As a default, the begining of the parameter list is intended to contain either a
   string, or
   the keyword :name followed by a string. If it is not the case, 'specialized-normalize' 
   considers itself as being not concerned with this parameter list, indicating this by 
   returning nil to 'normalize-compo'.
   If a string is present but not preceded by the keyword :name, this keyword is
   appended in front of the string in the result.
   In both cases, the result of applying 'normalize-compo' to whatever follows the name
   (using the same specializer, class and remaining argument list than received, is
   appended to this partial result."

  (cond
   ((stringp (car l))
    (list* :name (car l) (apply #'normalize-compo (cdr l) class spec defaults restargs) ) )
   ((eq (car l) :name)
    (list* :name (cadr l) (apply #'normalize-compo (cddr l) class spec defaults restargs) ) ) ) )