;;; A Common Lisp implementation of the compo protocol Version 2
;;;
;;; Package :compo
;;;
;;; Copyright (C) 2001  Bruno LARTILLOT
;;; 
;;; This program is free software; you can redistribute it and/or
;;; modify it under the terms of the GNU General Public License
;;; as published by the Free Software Foundation; either version 2
;;; of the License, or any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with this program; if not, write to the Free Software
;;; Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

(in-package :compo)

(defgeneric sub-print (sub super stream)
  (:documentation

"Generic Function
   sub-print (sub super stream)
Arguments
   A subcomposition to be printed.
   The name of the class of the composition that is the root of the structure currently
   being printed.
   The current output stream.
Values
   No value is returned.
   As a side effect, the subcomposition is printed in the given output stream." ) )

(defmethod sub-print (sub super stream)

"Primary method
   defaulting sub-print
Behaviour
   The object is printed in the output stream."

  (declare (ignore super))
  (print-object sub stream) )

(defmethod sub-print ((sub abstract-compo) super stream)

"Primary-method
   sub-print specialized on 'compo'
Behaviour
   An opening parenthesis starts the printing of the composition, and a closing
   parenthesis ends it.
   sub-print prints the class name only if its differs from the class name of the root,
   delegates the printing of the eventual properties defined by any subclass of 'compo'
   to the generic function 'specialized-print' and recursively applies sub-print for
   each subcomposition, taking care of the two following special cases :
   . If the value of the slot 'subs' in the composition is not a list, it represents the
     particular case of an atomic subcomposition, which is indicated by making its
     representation preceded by the keyword :subs.
   . In that case of an atomic subcomposition, if the subcomposition represents the
     composition itself, it is represented by the keyword :self to avoid any circularity
     when printing."

  (write-string "(" stream)
  (let ((sep))
    (unless (eq (class-name (class-of sub))
                (class-name (class-of super)) )
      (write (class-name (class-of sub)) :stream stream)
      (setf sep t) )
    (setf sep (specialized-print sub stream sep))
    (when (subs sub)
      (when sep (write-string " " stream))
      (if (eq (subs sub) sub)
        (write :self :stream stream)
        (if (atom (subs sub))
          (progn
            (write-string ". " stream)
            (sub-print (subs sub) super stream) )
          (mapl #'(lambda (x)
                    (sub-print (car x) super stream) )
                (subs sub) ) ) ) )
    (write-string ")" stream) ) )