;;; A Common Lisp implementation of the compo protocol Version 2
;;;
;;; Package :note
;;;
;;; Copyright (C) 2001  Bruno LARTILLOT
;;; 
;;; This program is free software; you can redistribute it and/or
;;; modify it under the terms of the GNU General Public License
;;; as published by the Free Software Foundation; either version 2
;;; of the License, or any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with this program; if not, write to the Free Software
;;; Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

(in-package :note)

(defclass abstract-note (abstract-compo)
  ((hei
    :initarg :hei
    :initform 0
    :accessor hei
    :type (or number list) )
   (pos
    :accessor pos
    :type (or number list) )
   (dur
    :initarg :dur
    :initform 1
    :accessor dur
    :type (or number list) )
   (voi
    :initarg :voi
    :initform 0
    :accessor voi
    :type (or number list) )
   (dyn
    :initarg :dyn
    :initform 1
    :accessor dyn
    :type (or number list) ) )
    (:documentation

"Class : 
   abstract-note
Direct Superclasses :
   abstract-compo
Slots :
   hei : of type number or list
     initarg : :hei
     accessor : hei
     defaults to 0
   pos : of type number or list
     accessor : pos
   dur : of type number or list
     initarg : :dur
     accessor : dur
     defaults to 1
   voi : of type number or list
     initarg : :voi
     accessor : voi
     defaults to 0
   dyn : of type number or list
     initarg : :dyn
     accessor : dyn
     defaults to 1
Purpose :
   The hei (height), voi (voice), pos (position), dur (duration) and dyn (dynamics)
   properties of notes always holds numeric expression which can be any number,
   function call or a combination of function calls which result is a number, or a
   quoted expression representing a coercion-specifier (see the context-item 'coercion'
   in the package :compo).
   For example : (+ (* 10 2) 4) or '(closestc 1 2 3) are valid values for any of these
   properties.
   In addition, the lexical variables super-hei, prev-hei, super-pos, prev-pos,
   super-dur, prev-dur, super-voi, prev-voi, super-dyn and prev-dyn can be
   used either as a simple value for these properties or as an argument to any of the
   valid functional or coercive expressions (see the 'evaluation' context item)." ) )

(defmethod initialize-instance ((self abstract-note) 
                                &key (pos 0)
                                     (pos-mod :abs)
                                     coerce-hei
                                     coerce-dur
                                     coerce-pos
                                     coerce-voi
                                     coerce-dyn
                                &allow-other-keys )

"Primary method
   initialize-instance specialized on 'abstract-note'
Behaviour
   The value slot :pos of the note being constructed is determined according to those
   of the two parameter keywords :pos-mod and :pos :
   . If :pos-mod is equal to :juxt, the position is list constituted of the expression
     (+ prev-pos prev-dur) and the value of the key argument :pos;
   . If :pos-mod is equal to :super, the position is list constituted of the expression
     (+ super-pos) and the value of the key argument :pos;
   . If :pos-mod is equal to :abs, the position is the value of the key argument :pos.
   :pos-mod defaults to :juxt and :pos defaults to 0.
   If any of the key arguments :coerce-hei :coerce-dur :coerce-pos :coerce-voi
   :coerce-dyn is provided, the slot 'coercer' from the superclass 'compo' is set to
   a new instance of the class 'note', whith each numeric property either initialized to
   the received value or defaulted to the form '(quote (identity))'."

  (call-next-method)
  (setf (pos self) (case pos-mod
                     ((:juxt) `((+ prev-pos prev-dur) ,pos))
                     ((:super) `((+ super-pos) ,pos))
                     ((:abs) pos) ))
  (when (or coerce-hei coerce-dur coerce-pos coerce-voi coerce-dyn)
    (setf (coercer self)
          (make-instance 'note
            :hei (or coerce-hei ''(identity))
            :dur (or coerce-dur ''(identity))
            :pos (or coerce-pos ''(identity))
            :voi (or coerce-voi ''(identity))
            :dyn (or coerce-dyn ''(identity)) ) ) )

  ; CLisp expects 'initialize-instance' to return 'self'
  self )