;;; A Common Lisp implementation of the compo protocol Version 2
;;;
;;; Package :note
;;;
;;; Copyright (C) 2001  Bruno LARTILLOT
;;; 
;;; This program is free software; you can redistribute it and/or
;;; modify it under the terms of the GNU General Public License
;;; as published by the Free Software Foundation; either version 2
;;; of the License, or any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with this program; if not, write to the Free Software
;;; Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

(in-package :note)

(defmethod transform ((note note) transformer)

"Primary method
   transform specialized on the class 'note
Behaviour
   Consider the transformation of an atomic numeric property according to the
   following rule :
   . 'hei' is augmented with 'super-hei' (i.e. hei is replaced by the text
     (+ hei super-hei))
   . 'pos' is multiplied by 'super-dur' (pos is replaced by the text (* pos super-dur))
   . 'dur' is multiplied by 'super-dur' (dur is replaced by the text (* dur super-dur))
   . 'voi' is augmented with 'super-voi' (voi is replaced by the text (+ voi super-voi))
   . 'dyn' is multiplied by 'super-dyn' (dyn is replaced by the text (* dyn super-dyn))
   where 'super-hei', 'super-voi', 'super-dur' and 'super-dyn' correspond to the slots
   'hei', 'voi', 'dur' and 'dyn' slots respectively in the supernote.
   Each numeric slot in the current note is transformed according to the following 
   rules :
   . If the slot is an atom, the rules above directly applies;
   . If the slot is a list, it is considered as either a functional call or a coercion
   specifier. In both cases, each argument (each item of the list except the first) is
   recursively replaced in the list by its transformation.
   The current note is returned."

  (labels ((f (a g)
             (if (atom a)
               (funcall g a)
               (if (eq (car a) 'quote)
                 (list 'quote (cons (caadr a) (mapcar #'(lambda (b) (f b g)) (cdadr a))))
                 (cons (car a) (mapcar #'(lambda (b) (f b g)) (cdr a))) ) ) ))
    
    (with-slots ((trans-hei hei)
                 (trans-dur dur)
                 (trans-voi voi)
                 (trans-dyn dyn) ) transformer
      (with-slots (hei voi pos dur dyn) note
        (setf hei (and hei (f hei #'(lambda (a) `(+ ,a ,trans-hei)))))
        (setf pos (f pos #'(lambda (a) `(* ,a ,trans-dur))))
        (setf dur (f dur #'(lambda (a) `(* ,a ,trans-dur))))
        (setf voi (f voi #'(lambda (a) `(+ ,a ,trans-voi))))
        (setf dyn (f dyn #'(lambda (a) `(* ,a ,trans-dyn)))) ) ) )
  note )